import sys
import re
import struct

op_names = ["MOVI",
            "MOVR",
            "GETM",
            "PUTM",
            "ADDI",
            "ADDR",
            "SUBI",
            "SUBR",
            "XORI",
            "XORR",
            "NOTI",
            "NOTR",
            "MULI",
            "MULR",
            "DIVI",
            "DIVR",
            "PUSH",
            "POOP",
            "CALL",
            "HALT",
            "NOPE"]

reg_names = ["R0", "R1", "R2", "R3", "S0", "S1", "S2", "S3", "IP", "BP", "SP"]
section_names = ["DATA:", "CODE:", "STACK:"]
section_flags = {s: i + 1 for i, s in enumerate(section_names)}
ops = {s: i for i, s in enumerate(op_names)}
regs = {s: i for i, s in enumerate(reg_names)}

assembled = bytearray()


def to_uint8(data):
    alphanum = re.compile("^[0-9]+$")
    if isinstance(data, int):
        return struct.pack("<b", data)
    elif data.startswith("0x"):
        return struct.pack("<b", int(data, 16))
    elif alphanum.match(data):  # only numbers
        return struct.pack("<b", int(data))
    else:
        return None

def to_uint16(data):
    print(data)
    alphanum = re.compile("^[0-9]+$")
    if isinstance(data, int):
        return struct.pack("<h", data)
    elif data.startswith("0x"):
        return struct.pack("<h", int(data, 16))
    elif alphanum.match(data):  # only numbers
        return struct.pack("<h", int(data))
    else:
        return None


def is_reg(data):
    if data not in reg_names:
        return False
    return True


def assemble_code(line):
    global assembled
    sys.stdout.write("CODE:")
    instruction = [x for x in re.split('\W', line) if x]
    op_name = instruction[0]

    if op_name not in op_names:
        sys.stderr.write(
            "ERROR WHILE ASSEMBLING UNKNOWN OPERATION: {}\n".format(op_name))
        return False
    if op_name == "MOVI":
        op = ops[op_name]
        dst = instruction[1]
        if (is_reg(dst)):
            dst = to_uint8(regs[dst])
            src = to_uint16(instruction[2])
            print("SRC: {} | DST: {}".format(src, dst))
        else:
            sys.stderr.write(
                "ERROR WHILE ASSEMBLING UNKNOWN REGISTER: {}\n".format(dst))
            return False
        assembled += to_uint8(op) + dst + src
    return True


def assemble_data(line):
    sys.stdout.write("DATA:\t")
    sys.stdout.write(line.strip(",") + "\n")


def main():
    global assembled
    with open("./test.gipu", 'r') as f:
        gen = (line.strip("\n") for line in f if line != "\n")

        for line in gen:
            if line.startswith(tuple(section_names)):
                flag = section_flags[line]
                continue
            if flag == section_flags["DATA:"]:
                assemble_data(line)
            elif flag == section_flags["CODE:"]:
                assemble_code(line)
    with open("./out.gipu", 'wb') as f:
        f.write(assembled)

if __name__ == '__main__':
    main()
